import importlib.resources
import re
import subprocess
import sys
import textwrap
from pathlib import Path

from wsgi_shim.wsgi_shim import cli_backend

from tests.conftest import gen_config_toml_contents
from tests.conftest import run_passenger_wsgi_py


# @pytest.mark.skip
def test_install_passenger_wsgi_py_django(tmp_path_world_readable, passenger_block):
    site_dir_path = tmp_path_world_readable
    cli_backend(argv=['wsgi-shim', 'install', str(site_dir_path)])
    passenger_app_root_path = site_dir_path / 'www-approot'

    # The command 'django-admin startproject mysite' is run in a subprocess
    # (rather than via django.core.management.execute_from_command_line) so
    # as not to configure settings before they are set up.
    subprocess.run(
        [Path(sys.executable).parent / 'django-admin', 'startproject', 'mysite'],
        cwd=site_dir_path,
    )
    mysite_path = site_dir_path / 'mysite'
    assert mysite_path.exists()
    subprocess.run(
        [sys.executable, 'manage.py', 'startapp', 'user'],
        cwd=mysite_path,
    )
    mysite_user_path = mysite_path / 'user'
    assert mysite_user_path.exists()
    models_py_path = mysite_user_path / 'models.py'
    models_py_path.write_text(
        textwrap.dedent("""
        from django.contrib.auth.models import AbstractUser

        class User(AbstractUser):
            pass
    """)[1:],
    )
    admin_py_path = mysite_user_path / 'admin.py'
    admin_py_path.write_text(
        textwrap.dedent("""
        from django.contrib import admin
        from django.contrib.auth.admin import UserAdmin
        from .models import User

        admin.site.register(User, UserAdmin)
    """)[1:],
    )
    settings_py_dir_path = mysite_path / 'mysite'
    settings_py_file_path = settings_py_dir_path / 'settings.py'

    with importlib.resources.files('tests').joinpath('django_local_settings.py').open('r') as f:
        local_settings_py_contents = f.read()
    local_settings_py_file_path = settings_py_dir_path / 'local_settings.py'
    local_settings_py_file_path.write_text(local_settings_py_contents)

    with open(settings_py_file_path, mode='a') as f:
        f.write(
            textwrap.dedent("""
        from mysite import local_settings
        SECRET_KEY = local_settings.update_secret_key(SECRET_KEY)
        DEBUG = local_settings.update_debug(DEBUG)
        local_settings.update_allowed_hosts(ALLOWED_HOSTS)
        local_settings.update_installed_apps(INSTALLED_APPS)
        local_settings.update_middleware(MIDDLEWARE)
        local_settings.update_databases(DATABASES)
        LOGGING = local_settings.get_logging()
        STATIC_ROOT = local_settings.get_static_root(BASE_DIR)
        AUTH_USER_MODEL = local_settings.get_auth_user_model()
        """),
        )
    log_dir_path = site_dir_path / 'log'
    log_dir_path.mkdir()
    log_file_path = log_dir_path / 'logfile'
    log_file_path.touch()
    secrets_dir_path = site_dir_path / 'secrets'
    secrets_dir_path.mkdir()
    credentials_file_path = secrets_dir_path / 'credentials.env'
    credentials_file_path.write_text(
        textwrap.dedent("""
    DJANGO_SECRET_KEY=__________placeholder_django_secret_key___________
    DB_NAME=placeholder_db_name
    DB_USER=placeholder_db_user
    DB_PASS=placeholder_db_pass
    DB_HOST=placeholder_db_host
    DB_PORT=3306
    """),
    )
    config_toml_path = passenger_app_root_path / 'config.toml'
    config_toml_contents = gen_config_toml_contents(
        {
            'passenger': passenger_block,
            'wsgi': {
                'chdir': str(mysite_path),
                'app': 'mysite.wsgi.application',
            },
            'secret_files': {
                'MYSITESECRETS': str(credentials_file_path),
            },
            'environment': {
                'MYSITEURL': 'placeholder',
                'LOG_FILENAME': str(log_file_path),
                'DJANGO_SETTINGS_MODULE': 'mysite.settings',
            },
        },
    )
    config_toml_path.write_text(config_toml_contents)
    restart_dir_path = passenger_app_root_path / 'tmp'
    maint_txt_path = restart_dir_path / 'maint.txt'
    maint_txt_path.unlink()
    html, status, headers, err = run_passenger_wsgi_py(passenger_app_root_path)
    log = log_file_path.read_text()
    assert err == []
    assert re.search(r'The install worked successfully', html)
    assert status == '200 OK'
    assert len(headers) == 6
    assert log == ''



# wsgi_shim.py
"""
Phusion Passenger WSGI shim

Ultimately, the only thing this really needs to do is to define
a callable called "application" roughly like this:

    from WSGI_MODULE import WSGI_APP as application

However, because we assume users do not have direct access to the server logs
to debug their code, this bit of code (and the accompanying documentation)
helps catch many initial setup/configuration errors without users needing
to ask sysadmins to find snippets in the server error logs.

This is not WSGI middleware; it does not capture exceptions thrown by the
application.  It is up to the developer to log messages to an accessible
file to aid in further debugging and to gather site statistics.
(This could be a future enhancement for this project, though.)
"""
import argparse
import datetime
import functools
import grp
import importlib.resources
import importlib.util
import os
import pwd
import re
import sys
import textwrap
import tomllib
from contextlib import contextmanager
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from types import ModuleType
from typing import cast
from typing import NoReturn
from typing import Protocol
from typing import Self
from wsgiref.types import StartResponse
from wsgiref.types import WSGIApplication
from wsgiref.types import WSGIEnvironment


class WSGIConfigException(Exception):
    pass


def is_safe_gid(file_gid: int) -> bool:
    """Returns False (indicating an unsafe file_gid) when the file_gid
    matches the user's default group and that default group's name is
    not that same string as the username."""
    passwd_entry = pwd.getpwuid(os.getuid())
    if passwd_entry.pw_gid == file_gid:
        if passwd_entry.pw_name != grp.getgrgid(passwd_entry.pw_gid).gr_name:
            return False
    return True


def is_path_world_readable(path: Path) -> bool:
    if not path.exists():
        raise WSGIConfigException(f"does not exist: '{path}'")
    p = Path('/')
    for part in path.parts:
        p = p / part
        mode = 0o001 if p.is_dir() else 0o004
        if (p.stat().st_mode & mode) != mode:
            return False
    return True


def check_path_is_world_readable(path: Path) -> None:
    if not is_path_world_readable(path):
        raise WSGIConfigException(f"not world readable: '{path}'")


def check_path_is_not_world_readable(path: Path) -> None:
    if is_path_world_readable(path):
        raise WSGIConfigException(f"is world readable: '{path}'")


def is_path_world_or_unsafe_gid_writable(path: Path) -> bool:
    if not path.exists():
        raise WSGIConfigException(f"does not exist: '{path}'")
    p = Path('/')
    for part in path.parts:
        p = p / part
        p_stat = p.stat()
        mode = 0o003 if p.is_dir() else 0o002
        if (p_stat.st_mode & mode) == mode:
            return True
        if not is_safe_gid(p_stat.st_gid):
            mode = 0o030 if p.is_dir() else 0o020
            if (p.stat().st_mode & mode) == mode:
                return True
    return False


def check_path_is_not_world_or_unsafe_gid_writable(path: Path) -> None:
    if is_path_world_or_unsafe_gid_writable(path):
        raise WSGIConfigException(f"is world (or unsafe group) writable: '{path}'")


def check_venv_path(passenger_python: Path, passenger_app_root: Path):
    if sys.prefix == sys.base_prefix:
        raise WSGIConfigException('Must be run from inside a virtual environment')
    passenger_venv_path = passenger_python.parent.parent
    running_venv_path = Path(sys.prefix)
    if running_venv_path != passenger_venv_path:
        raise WSGIConfigException(
            'Running virtual environment does not match '
            'passenger_python in config.toml.',
        )
    if running_venv_path.is_relative_to(passenger_app_root):
        raise WSGIConfigException(
            'Virtual environment cannot be '
            'inside passenger_app_root',
        )
    venv_stat = os.stat(running_venv_path)
    venv_mask = 0o002 if is_safe_gid(venv_stat.st_gid) else 0o022
    if (venv_stat.st_mode & venv_mask) != 0:
        raise WSGIConfigException(
            f'virtual environment at {running_venv_path} '
            f'is not adequately write protected.',
        )


@dataclass
class Config:
    passenger_user: str = ''
    passenger_group: str = ''
    passenger_python: str = ''
    wsgi_chdir: str = ''
    wsgi_app: str = ''
    environment: dict[str, tuple[str, bool]] = field(default_factory=dict)
    maintenance_mode_path: Path = Path('/dev/null/nonexistent')
    passenger_python_path: Path = Path('/dev/null/nonexistent')

    @classmethod
    def from_toml_dict(cls, data, passenger_app_root: Path) -> Self:
        retval = cls()
        for section, required in [
            ('passenger', True),
            ('wsgi', False),
            ('secret_files', False),
            ('environment', False),
        ]:
            if not isinstance(data.get(section, {}), dict):
                raise WSGIConfigException(f'"{section}" is not a section')
            if required and section not in data:
                raise WSGIConfigException(f'Missing required [{section}]')
        for section, key, attrib in [
            ('passenger', 'passenger_user', 'passenger_user'),
            ('passenger', 'passenger_group', 'passenger_group'),
            ('passenger', 'passenger_python', 'passenger_python'),
            ('wsgi', 'chdir', 'wsgi_chdir'),
            ('wsgi', 'app', 'wsgi_app'),
        ]:
            value = data.get(section, {}).get(key, '')
            if not isinstance(value, str):
                raise WSGIConfigException(f'{section}.{key} not a string')
            setattr(retval, attrib, value)
        retval.maintenance_mode_path = passenger_app_root / 'tmp' / 'maint.txt'
        retval.passenger_python_path = Path(retval.passenger_python)
        environment_sections = [('secret_files', True), ('environment', False)]
        for section, is_secret_file in environment_sections:
            for key, value in data.get(section, {}).items():
                if re.fullmatch(r'[a-zA-Z_][a-zA-Z0-9_]*', key) is None:
                    raise WSGIConfigException(f'invalid key "{key}" in [{section}]')
                if not isinstance(value, str):
                    raise WSGIConfigException(f'{section}.{key} not a string')
                if key in retval.environment:
                    raise WSGIConfigException(
                        f'key "{key}" duplicated across sections: '
                        f'{", ".join(name for name, _ in environment_sections)}',
                    )
                retval.environment[key] = (value, is_secret_file)
        return retval

    def check_user_group(self):
        running_user = pwd.getpwuid(os.getuid()).pw_name
        running_group = grp.getgrgid(os.getgid()).gr_name
        running = (running_user, running_group)
        configured = (self.passenger_user, self.passenger_group)
        if running != configured:
            raise WSGIConfigException(
                f'running user.group {running} does not '
                f'match config.toml settings {configured}.',
            )

    def check_secret_files(self):
        # Check that the secret_files exist and have appropriate permissions.
        # Note: Like ssh, we refuse to run if a secrets file is too widely
        # accessible.
        for key, (filename, is_secret_file) in self.environment.items():
            if is_secret_file:
                path = Path(filename)
                if not path.is_absolute():
                    raise WSGIConfigException(
                        f'secret_files.{key} '
                        f'is not an absolute path',
                    )
                if not os.access(path, os.R_OK):
                    raise WSGIConfigException(
                        f'secret_files.{key} '
                        f'is missing or not readable',
                    )
                config_stat = os.stat(path)
                config_mask = 0o007 if is_safe_gid(config_stat.st_gid) else 0o077
                if (config_stat.st_mode & config_mask) != 0:
                    raise WSGIConfigException(
                        f'secret_files.{key} '
                        f'is not adequately protected.',
                    )

    def update_os_environment(self) -> None:
        for key, (value, _) in self.environment.items():
            if key in os.environ:
                raise WSGIConfigException(f'os.environ already has key "{key}"')
            os.environ[key] = value

    def in_maintenance_mode(self) -> bool:
        return self.maintenance_mode_path.exists()


def load_config(file: Path):
    try:
        with open(file, "rb") as f:
            data = tomllib.load(f)  # assumes utf-8 encoding
    except FileNotFoundError as e:
        raise WSGIConfigException(f'Config file not found: {str(e)}')
    except PermissionError as e:
        raise WSGIConfigException(f'Config file error: {str(e)}')
    except tomllib.TOMLDecodeError as e:
        raise WSGIConfigException(f'Config file syntax error: {str(e)}')
    return data


def import_module_from_file_location(
        module_path: Path,
) -> ModuleType:
    # See: https://docs.python.org/3.11/library/importlib.html
    #      #importing-a-source-file-directly
    module_name = module_path.stem
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    if spec is None:
        raise WSGIConfigException(
            f'Cannot load {module_name} from {module_path}: '
            f'spec_from_file_location returned None.',
        )
    importlib.invalidate_caches()
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    assert spec.loader is not None  # keep mypy happy
    spec.loader.exec_module(module)
    return module


def application503(
        _environ: WSGIEnvironment,
        start_response: StartResponse,
        html: str,
) -> list[bytes]:
    html_as_bytes = html.encode('utf-8')
    start_response(
        '503 Service Unavailable',
        [
            ('Content-Type', 'text/html; charset=utf-8'),
            ('Content-Length', str(len(html_as_bytes))),
            ('Retry-After', '60'),
        ],
    )
    return [html_as_bytes]


html_template = textwrap.dedent('''\
    <!DOCTYPE html>
    <html lang="en">
    <head>
    <meta charset="utf-8"/>
    <title>{title} (503)</title>
    </head>
    <body>
    <h1>{title}</h1>
    <p>{reason}</p>
    {details}</body>
    </html>
''')


def get_app(passenger_app_root: Path | None = None) -> WSGIApplication:
    """
    Returns the appropriate application callable to use.
    """
    try:
        if passenger_app_root is None:  # pragma: no cover
            passenger_app_root = Path.cwd()
        config_file = passenger_app_root / 'config.toml'
        toml_dict = load_config(config_file)
        config = Config.from_toml_dict(toml_dict, passenger_app_root)
        check_venv_path(config.passenger_python_path, passenger_app_root)
        config.check_user_group()
        if config.in_maintenance_mode():
            return cast(
                WSGIApplication,
                functools.partial(
                    application503,
                    html=html_template.format(
                        title='Maintenance',
                        reason='The developer of this site has put it into '
                               'maintenance mode.  Please try again later.',
                        details=f'<p>{datetime.datetime.now()}</p>',
                    ),
                ),
            )
        config.check_secret_files()
        config.update_os_environment()
        # See: https://docs.python.org/3.11/library/importlib.html
        #      #importing-programmatically
        chdir_path = Path.cwd() / Path(config.wsgi_chdir)
        if not chdir_path.is_dir():
            raise WSGIConfigException(
                f'[wsgi].chdir "{chdir_path}" must be a directory',
            )
        s = config.wsgi_app.rsplit('.', maxsplit=1)
        if len(s) != 2:
            raise WSGIConfigException(
                '[wsgi].app must of the form module.app_name',
            )
        submodule_name, app_name = s
        # Note: doing a chdir is generally frowned upon but Django needs it
        # (otherwise, we'd just add chdir_path to sys.path).
        os.chdir(chdir_path)
        for d in (str(chdir_path), ''):
            while d in sys.path:
                sys.path.remove(d)
        sys.path.insert(0, str(chdir_path))
        try:
            # FIXME: verify os.environ and cwd
            module = importlib.import_module(submodule_name)
            app = getattr(module, app_name)
            return app
        except (ModuleNotFoundError, PermissionError, AttributeError) as exc_info:
            raise WSGIConfigException(
                f'Cannot import {config.wsgi_app} from {config.wsgi_chdir}: '
                f'{str(exc_info)}',
            )
    except WSGIConfigException as exc_info:
        # Configuration Error Mode
        return cast(
            WSGIApplication,
            functools.partial(
                application503,
                html=html_template.format(
                    title='Error Page',
                    reason='This site has experienced a configuration '
                           'error or exception.',
                    details=textwrap.dedent(f'''\
                        <pre>
                        passenger_app_root: {passenger_app_root}
                               process uid: {pwd.getpwuid(os.getuid()).pw_name}
                               process gid: {grp.getgrgid(os.getgid()).gr_name}
                                 timestamp: {datetime.datetime.now()}

                        Exception Details
                        {{}}
                        </pre>
                    ''').format(str(exc_info)),
                ),
            ),
        )


class HasExitError(Protocol):
    def exit(self, status=0, message=None) -> NoReturn:
        ...  # pragma no cover

    def error(self, message) -> NoReturn:
        ...  # pragma no cover


@dataclass
class Args:
    command: str
    sitedir: str
    approot: str
    docroot: str
    verbose: bool
    puppet: bool
    site_path: Path
    approot_path: Path
    docroot_path: Path
    parser: HasExitError


def cli_parse(argv=None) -> Args:
    if argv is None:
        argv = sys.argv
    args = argv[1:]
    parser = argparse.ArgumentParser(
        description="Create/check directory structure for Passenger Python web app",
    )
    parser.add_argument('command', choices=['install', 'check'])
    parser.add_argument(
        '-a', '--approot', default='www-approot',
        help='AppRoot relative to sitedir',
    )
    parser.add_argument(
        '-d', '--docroot', default='www-docroot',
        help='DocRoot relative to sitedir',
    )
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument(
        '-p', '--puppet', action='store_true',
        help='output puppet configuration',
    )
    parser.add_argument(
        'sitedir',
        help='absolute path of the site directory',
    )
    args = parser.parse_args(args=args)
    site_path = Path(args.sitedir)
    if not site_path.is_absolute():
        parser.error('sitedir must start with a "/".')
    approot_rel_path = Path(args.approot)
    docroot_rel_path = Path(args.docroot)
    for rel_path, name in [(approot_rel_path, 'AppRoot'), (docroot_rel_path, 'DocRoot')]:
        if rel_path.is_absolute():
            parser.error(f'{name} must be relative (no leading "/").')
        if len(rel_path.parts) != 1:
            parser.error(f'{name} can only be a single part (no internal "/").')
        if rel_path.parts[0] in ('.', '..'):
            parser.error(f'{name} cannot be "." or ".."')
    args.site_path = site_path
    args.approot_path = site_path / approot_rel_path
    args.docroot_path = site_path / docroot_rel_path
    args.parser = parser
    return cast(Args, args)


@contextmanager
def temp_umask(mask: int):
    old_mask = os.umask(mask)
    try:
        yield
    finally:
        os.umask(old_mask)


def get_umask() -> int:
    mask = os.umask(0o777)  # to get the umask, we must change it
    os.umask(mask)  # change it right back
    return mask


def cleanup(items: list[Path]) -> None:
    for item in reversed(items):
        if item.is_dir():
            try:
                item.rmdir()
            except OSError as exc_info:
                if exc_info.strerror != 'Directory not empty':
                    raise
        else:
            item.unlink(missing_ok=True)


def create_new_directory(path: Path, other_readable: bool = False) -> None:
    # create with no group/other permissions
    with temp_umask(0o077):
        path.mkdir()
    # now that the directory exists, check group
    path_stat = os.stat(path)
    current_mode = path_stat.st_mode
    new_mode = current_mode
    if other_readable:
        new_mode |= 0o005  # o=rx
    if is_safe_gid(path_stat.st_gid):
        # group is safe, update group bits to what they would have been based on umask
        new_mode |= 0o070 & ~get_umask()
    if new_mode != current_mode:
        path.chmod(new_mode)


def create_new_file(path: Path, content: str = '', other_readable: bool = False) -> None:
    # create with no group/other permissions
    with temp_umask(0o077):
        with open(path, mode='x') as f:
            f.write(content)
    # now that the directory exists, check group
    path_stat = os.stat(path)
    current_mode = path_stat.st_mode
    new_mode = current_mode
    if other_readable:
        new_mode |= 0o004  # o=r
    if is_safe_gid(path_stat.st_gid):
        # group is safe, update group bits to what they would have been based on umask
        new_mode |= 0o060 & ~get_umask()
    if new_mode != current_mode:
        path.chmod(new_mode)


def puppet_block(args: Args) -> str:
    return textwrap.dedent(f"""
        Requesting these attributes:
        passenger_enabled     => true,
        passenger_user        => '{args.approot_path.owner()}',
        passenger_group       => '{args.approot_path.group()}',
        passenger_app_root    => '{args.approot_path}',
        passenger_restart_dir => '{args.approot_path / 'tmp'}',
        passenger_app_type    => 'wsgi',
        passenger_python      => '{Path(sys.prefix) / 'bin' / 'python'}';
        """)


def cli_install(args: Args) -> None:
    # Keep track of what we created to know what to clean-up in case of error
    created_paths: list[Path] = list()
    # Create the directories
    restart_dir_path = args.approot_path / 'tmp'
    for dir_path in [args.docroot_path, args.approot_path, restart_dir_path]:
        try:
            create_new_directory(dir_path, other_readable=True)
        except FileExistsError:
            cleanup(created_paths)
            raise WSGIConfigException(
                f"Error: will not overwrite existing: "
                f"'{dir_path}'",
            )
        created_paths.append(dir_path)
    # File paths and contents
    passenger_wsgi_py_path = args.approot_path / 'passenger_wsgi.py'
    passenger_wsgi_py = textwrap.dedent('''
        # Created with wsgi_shim
        import wsgi_shim

        application = wsgi_shim.get_app()
    ''')[1:]
    config_toml_path = args.approot_path / 'config.toml'
    with importlib.resources.files('wsgi_shim.templates').joinpath(
            'config.toml',
    ).open('r') as f:
        config_template = f.read()
    config_toml = config_template.format(
        web_doc_root=args.docroot_path,
        user=args.approot_path.owner(),
        group=args.approot_path.group(),
        passenger_app_root=args.approot_path,
        passenger_restart_dir=restart_dir_path,
        passenger_python=Path(sys.prefix) / 'bin' / 'python',
    )
    restart_txt_path = restart_dir_path / 'restart.txt'
    maint_txt_path = restart_dir_path / 'maint.txt'
    # create the files
    for path, content in [
        (passenger_wsgi_py_path, passenger_wsgi_py),
        (config_toml_path, config_toml),
        (restart_txt_path, ''),
        (maint_txt_path, ''),
    ]:
        try:
            create_new_file(
                path=path,
                content=content,
                other_readable=False,
            )
        except FileExistsError:
            cleanup(created_paths)
            raise WSGIConfigException(
                f"Error: will not overwrite existing: "
                f"'{path}'",
            )
        created_paths.append(path)
    if args.verbose:
        print('Success.  Created the following:')
        for path in created_paths:
            print(f'    {path}')
    if args.puppet:
        print(puppet_block(args))


def cli_check(args: Args) -> None:
    passenger_wsgi_py_path = args.approot_path / 'passenger_wsgi.py'
    check_path_is_not_world_readable(passenger_wsgi_py_path)
    if not passenger_wsgi_py_path.is_file():
        raise WSGIConfigException(f"not a file: '{passenger_wsgi_py_path}'")
    contents = passenger_wsgi_py_path.read_text()
    try:
        compile(contents, str(passenger_wsgi_py_path), mode='exec')
    except (SyntaxError, ValueError) as exc_info:
        raise WSGIConfigException(f"Cannot parse: '{passenger_wsgi_py_path}': {exc_info}")
    except IsADirectoryError:
        raise WSGIConfigException(f"not a file: '{passenger_wsgi_py_path}'")
    restart_dir_path = args.approot_path / 'tmp'
    check_path_is_world_readable(restart_dir_path)
    if not restart_dir_path.is_dir():
        raise WSGIConfigException(f"not a directory: '{restart_dir_path}'")
    restart_txt_path = restart_dir_path / 'restart.txt'
    if restart_txt_path.exists():
        check_path_is_not_world_readable(restart_txt_path)
        if not restart_txt_path.is_file():
            raise WSGIConfigException(f"not a file: '{restart_txt_path}'")
    maint_txt_path = restart_dir_path / 'maint.txt'
    if maint_txt_path.exists():
        check_path_is_not_world_readable(maint_txt_path)
        if not maint_txt_path.is_file():
            raise WSGIConfigException(f"not a file: '{maint_txt_path}'")
    config_toml_path = args.approot_path / 'config.toml'
    check_path_is_not_world_readable(config_toml_path)
    toml_dict = load_config(config_toml_path)
    config = Config.from_toml_dict(toml_dict, args.approot_path)
    check_venv_path(config.passenger_python_path, args.approot_path)
    config.check_secret_files()
    if args.verbose:
        print('Check passed')
    if args.puppet:
        print(puppet_block(args))


def cli_backend(argv=None) -> None:
    args = cli_parse(argv)
    check_path_is_world_readable(args.site_path)
    if not args.site_path.is_dir():
        raise WSGIConfigException(f'not a directory: {args.site_path}')
    running_python_path = Path(sys.prefix) / 'bin' / 'python'
    check_venv_path(running_python_path, args.approot_path)
    if args.command == 'install':
        cli_install(args)
    elif args.command == 'check':
        cli_check(args)


def cli():  # pragma no cover
    try:
        cli_backend()
        return_code = 0
    except WSGIConfigException as exc_info:
        print(f'{str(exc_info)}', file=sys.stderr)
        return_code = 1
    raise SystemExit(return_code)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #



import importlib.resources
import importlib.util
import os
import re
import sys
import textwrap
from dataclasses import dataclass
from pathlib import Path
from unittest import mock

import pytest
from wsgi_shim import check_path_is_not_world_readable
from wsgi_shim import check_path_is_world_readable
from wsgi_shim import Config
from wsgi_shim import get_app
from wsgi_shim import is_path_world_readable
from wsgi_shim import load_config
from wsgi_shim import WSGIConfigException
from wsgi_shim.wsgi_shim import check_venv_path
from wsgi_shim.wsgi_shim import cli_backend
from wsgi_shim.wsgi_shim import cli_parse
from wsgi_shim.wsgi_shim import create_new_file
from wsgi_shim.wsgi_shim import is_path_world_or_unsafe_gid_writable
from wsgi_shim.wsgi_shim import is_safe_gid

from tests.conftest import gen_config_toml_contents
from tests.conftest import run_app
from tests.conftest import run_passenger_wsgi_py


def test_safe_gid_false(monkeypatch):
    @dataclass
    class PasswordEntry:
        pw_gid: int = 9999
        pw_name: str = 'username'

    @dataclass
    class GroupEntry:
        gr_name: str = 'groupname'

    monkeypatch.setattr('os.getuid', lambda: 9999)
    monkeypatch.setattr('pwd.getpwuid', lambda _: PasswordEntry())
    monkeypatch.setattr('grp.getgrgid', lambda _: GroupEntry())
    assert not is_safe_gid(9999)


@pytest.mark.parametrize(
    'uid,gid,uname,gname,file_gid',
    [
        pytest.param(
            9999, 8888, 'username', 'groupname', 1111, id='differ_gid',
        ),
        pytest.param(
            9999, 8888, 'name', 'name', 8888, id='same_name',
        ),
    ],
)
def test_safe_gid_true(uid, gid, uname, gname, file_gid, monkeypatch):
    @dataclass
    class PasswordEntry:
        pw_gid: int = gid
        pw_name: str = uname

    @dataclass
    class GroupEntry:
        gr_name: str = gname

    monkeypatch.setattr('os.getuid', lambda: uid)
    monkeypatch.setattr('pwd.getpwuid', lambda _: PasswordEntry())
    monkeypatch.setattr('grp.getgrgid', lambda _: GroupEntry())
    assert is_safe_gid(file_gid)  # noqa


def test_is_path_world_readable_no_file():
    with pytest.raises(WSGIConfigException, match=r'does not exist'):
        is_path_world_readable(Path('/dev/null/not_a_file'))


def test_is_path_world_readable_true(tmp_path_world_readable):
    file_path = tmp_path_world_readable / 'foo'
    create_new_file(file_path, other_readable=True)
    assert is_path_world_readable(file_path)


def test_check_path_is_world_readable_fail():
    file = Path('/root')
    with pytest.raises(WSGIConfigException, match=r'not world readable'):
        check_path_is_world_readable(file)


def test_check_path_is_not_world_readable_fail():
    file = Path('/')
    with pytest.raises(WSGIConfigException, match=r'is world readable'):
        check_path_is_not_world_readable(file)


def test_is_path_world_or_unsafe_gid_writable_no_path():
    with pytest.raises(WSGIConfigException, match=r'does not exist'):
        is_path_world_or_unsafe_gid_writable(Path('/dev/null/no_exist'))


def test_is_path_world_or_unsafe_gid_writable_file(tmp_path_world_readable):
    tmp_path_world_readable.chmod(tmp_path_world_readable.stat().st_mode | 0o003)
    file_path = tmp_path_world_readable / 'foo'
    create_new_file(file_path)
    file_path.chmod(file_path.stat().st_mode | 0o002)
    assert is_path_world_or_unsafe_gid_writable(file_path)


def test_is_path_world_or_unsafe_gid_writable_dir(tmp_path_world_readable):
    tmp_path_world_readable.chmod(tmp_path_world_readable.stat().st_mode | 0o003)
    dir_path = tmp_path_world_readable / 'foo'
    dir_path.mkdir()
    dir_path.chmod(dir_path.stat().st_mode | 0o003)
    assert is_path_world_or_unsafe_gid_writable(dir_path)


def test_check_venv_path_not_in_venv(monkeypatch):
    monkeypatch.setattr('sys.base_prefix', sys.prefix)
    with pytest.raises(WSGIConfigException, match=r'from inside a virtual'):
        check_venv_path(Path(), Path())


def test_check_venv_path_mismatch():
    with pytest.raises(WSGIConfigException, match=r'does not match'):
        check_venv_path(
            passenger_python=Path('/dev/null/nonexistent/nonexistent'),
            passenger_app_root=Path(),
        )


def test_check_venv_path_error_inside():
    with pytest.raises(WSGIConfigException, match=r'cannot be inside'):
        check_venv_path(
            passenger_python=Path(sys.prefix) / 'bin' / 'python',
            passenger_app_root=Path('/'),
        )


def test_load_config_empty():
    path = Path('/dev/null')
    data = load_config(path)
    assert isinstance(data, dict)
    assert len(data) == 0


def test_load_config_valid_1(tmp_path):
    file = tmp_path / 'config.toml'
    file.write_text("""
    [wsgi]
    """)
    data = load_config(file)
    assert isinstance(data, dict)
    assert data == {'wsgi': {}}


def test_load_config_fail_perms(tmp_path):
    file = tmp_path / 'config.toml'
    file.write_text("""
    [wsgi]
    """)
    file.chmod(0o200)
    with pytest.raises(WSGIConfigException, match=r'Permission denied'):
        load_config(file)


def test_load_config_fail_missing(tmp_path):
    file = tmp_path / 'config.toml'
    with pytest.raises(WSGIConfigException, match=r'file not found'):
        load_config(file)


def test_load_config_fail_toml_syntax(tmp_path):
    file = tmp_path / 'config.toml'
    file.write_text("""
    [
    """)
    with pytest.raises(WSGIConfigException, match=r'syntax error'):
        load_config(file)


def test_load_config_fail_toml_duplicated_section(tmp_path):
    file = tmp_path / 'config.toml'
    file.write_text("""
    [wsgi]
    [wsgi]
    """)
    with pytest.raises(WSGIConfigException, match=r'syntax error'):
        load_config(file)


@pytest.mark.parametrize(
    'section',
    [
        'wsgi',
        'secret_files',
        'environment',
    ],
)
def test_load_config_fail_toml_not_section(section, tmp_path):
    file = tmp_path / 'config.toml'
    file.write_text(f"""
    {section} = "foo"
    [passenger]
    """)
    with pytest.raises(WSGIConfigException, match=r'not a section'):
        data = load_config(file)
        Config.from_toml_dict(data, tmp_path)


@pytest.mark.parametrize(
    'section,key',
    [
        ('wsgi', 'chdir'),
        ('wsgi', 'app'),
    ],
)
def test_load_config_fail_toml_values_not_str_1(section, key):
    data = {section: {key: 1}}
    if section != 'passenger':
        data['passenger'] = {}
    with pytest.raises(WSGIConfigException, match=fr'{section}.{key} not a string'):
        Config.from_toml_dict(data, Path())


@pytest.mark.parametrize(
    'section,key',
    [
        ('secret_files', 'filename'),
        ('environment', 'foobar'),
    ],
)
def test_load_config_fail_toml_values_not_str(section, key):
    data = {'passenger': {}, section: {key: 1}}
    with pytest.raises(WSGIConfigException, match=fr'{section}.{key} not a string'):
        Config.from_toml_dict(data, Path())


@pytest.mark.parametrize(
    'section,key',
    [
        ('secret_files', '0_key_leading_digit'),
        ('environment', 'key_$_special_char'),
    ],
)
def test_load_config_fail_toml_invalid_key(section, key):
    data = {'passenger': {}, section: {key: 1}}
    with pytest.raises(WSGIConfigException, match=fr'invalid key .* in \[{section}\]'):
        Config.from_toml_dict(data, Path())


def test_load_config_fail_toml_duplicate_keys():
    data = {
        'passenger': {},
        'wsgi': {},
        'secret_files': {'key': 'value1'},
        'environment': {'key': 'value2'},
    }
    with pytest.raises(WSGIConfigException, match=r'duplicated across sections'):
        Config.from_toml_dict(data, Path())


def test_update_os_environment_conflict(monkeypatch):
    data = {
        'passenger': {},
        'environment': {'FOOBAR': 'value'},
    }
    monkeypatch.setenv('FOOBAR', 'something')
    with pytest.raises(WSGIConfigException, match=r'already has key'):
        config = Config.from_toml_dict(data, Path())
        config.update_os_environment()


@mock.patch.dict(os.environ, {})
def test_update_os_environment_success():
    data = {
        'passenger': {},
        'environment': {'FOOBAR': 'value'},
    }
    config = Config.from_toml_dict(data, Path())
    assert 'FOOBAR' not in os.environ
    config.update_os_environment()
    assert os.environ['FOOBAR'] == 'value'


def test_check_secret_files_fail_missing(tmp_path):
    missing_file = tmp_path / 'not_there'
    data = {
        'passenger': {},
        'secret_files': {'SECRET': str(missing_file)},
    }
    config = Config.from_toml_dict(data, Path())
    with pytest.raises(WSGIConfigException, match=r'is missing or not readable'):
        config.check_secret_files()


def test_check_secret_files_fail_cant_read(tmp_path):
    unreadable_file = tmp_path / 'unreadable'
    unreadable_file.touch(mode=0o000)
    data = {
        'passenger': {},
        'secret_files': {'SECRET': str(unreadable_file)},
    }
    config = Config.from_toml_dict(data, Path())
    with pytest.raises(WSGIConfigException, match=r'is missing or not readable'):
        config.check_secret_files()


def test_check_secret_files_fail_too_permissive(tmp_path):
    permissive_file = tmp_path / 'permissive'
    permissive_file.touch()
    permissive_file.chmod(mode=0x777)
    data = {
        'passenger': {},
        'secret_files': {'SECRET': str(permissive_file)},
    }
    config = Config.from_toml_dict(data, Path())
    with pytest.raises(WSGIConfigException, match=r'not adequately protected'):
        config.check_secret_files()


def test_check_secret_files_fail_not_absolute():
    data = {
        'passenger': {},
        'secret_files': {'FOOBAR': 'not_absolute_path'},
    }
    config = Config.from_toml_dict(data, Path())
    with pytest.raises(WSGIConfigException, match=r'not an absolute path'):
        config.check_secret_files()


def test_hello_world():
    from src.wsgi_shim.hello_world import app_hello_world as app_under_test
    html, status, headers, err = run_app(app_under_test)
    assert err == []
    assert re.search(r'Hello, World', html)
    assert status == '200 OK'
    assert len(headers) == 2


def test_main_maintenance_mode_normal(tmp_path, passenger_block):
    passenger_tmp_dir = tmp_path / 'tmp'
    passenger_tmp_dir.mkdir(mode=0o755)
    maint_file = passenger_tmp_dir / 'maint.txt'
    maint_file.touch()
    config_file = tmp_path / 'config.toml'
    config_file.write_text(f"""
    [passenger]
    {passenger_block}
    """)
    app = get_app(tmp_path)
    html, status, headers, err = run_app(app)
    assert err == []
    assert re.search(r'put it into maintenance mode', html)
    assert status == '503 Service Unavailable'
    assert len(headers) == 3


def test_main_fail_venv_mismatch(tmp_path):
    config_file = tmp_path / 'config.toml'
    config_file.write_text(f"""
    [passenger]
    passenger_python="{'foo'}"
    [wsgi]
    chdir = "not_there"
    app = "unused"
    """)
    app = get_app(tmp_path)
    html, status, headers, err = run_app(app)
    assert err == []
    assert re.search(r'does not match passenger_python', html)
    assert status == '503 Service Unavailable'
    assert len(headers) == 3


def test_main_fail_no_chdir(tmp_path_to_scaffold, passenger_block):
    site_dir_path = tmp_path_to_scaffold
    approot_path = site_dir_path / 'www-approot'
    config_toml_path = approot_path / 'config.toml'
    config_toml_contents = gen_config_toml_contents(
        {
            'passenger': passenger_block,
            'wsgi': {
                'chdir': 'not_there',
                'app': 'unreadable.whatever',
            },
        },
    )
    config_toml_path.write_text(config_toml_contents)
    app = get_app(approot_path)
    html, status, headers, err = run_app(app)
    assert err == []
    assert re.search(r'not_there" must be a directory', html)
    assert status == '503 Service Unavailable'
    assert len(headers) == 3


def test_main_fail_module_perms(tmp_path_to_scaffold, passenger_block):
    site_dir_path = tmp_path_to_scaffold
    mysite_path = site_dir_path / 'mysite'
    mysite_path.mkdir()
    unreadable_file = mysite_path / 'unreadable.py'
    unreadable_file.touch(mode=0o000)
    approot_path = site_dir_path / 'www-approot'
    config_toml_path = approot_path / 'config.toml'
    config_toml_contents = gen_config_toml_contents(
        {
            'passenger': passenger_block,
            'wsgi': {
                'chdir': str(mysite_path),
                'app': 'unreadable.whatever',
            },
        },
    )
    config_toml_path.write_text(config_toml_contents)
    app = get_app(approot_path)
    html, status, headers, err = run_app(app)
    assert err == []
    assert re.search(r'Cannot import .* Permission denied: ', html)
    assert status == '503 Service Unavailable'
    assert len(headers) == 3


def test_main_fail_no_app(tmp_path_to_scaffold, passenger_block):
    site_dir_path = tmp_path_to_scaffold
    mysite_path = site_dir_path / 'mysite'
    mysite_path.mkdir()

    empty_file = mysite_path / 'empty_file.py'
    empty_file.touch()

    approot_path = site_dir_path / 'www-approot'
    config_toml_path = approot_path / 'config.toml'
    config_toml_contents = gen_config_toml_contents(
        {
            'passenger': passenger_block,
            'wsgi': {
                'chdir': str(mysite_path),
                'app': 'empty_file.whatever',
            },
        },
    )
    config_toml_path.write_text(config_toml_contents)
    app = get_app(approot_path)
    html, status, headers, err = run_app(app)
    assert err == []
    assert re.search(r"module 'empty_file' has no attribute 'whatever'", html)
    assert status == '503 Service Unavailable'
    assert len(headers) == 3


def test_main_normal_chdir_file(tmp_path_to_scaffold, passenger_block):
    site_dir_path = tmp_path_to_scaffold
    mysite_path = site_dir_path / 'mysite'
    mysite_path.mkdir()
    app_module_file = mysite_path / 'myapp.py'
    with importlib.resources.files('wsgi_shim').joinpath('hello_world.py').open('r') as f:
        app_source = f.read()
    app_module_file.write_text(app_source)
    approot_path = site_dir_path / 'www-approot'
    config_toml_path = approot_path / 'config.toml'
    config_toml_contents = gen_config_toml_contents(
        {
            'passenger': passenger_block,
            'wsgi': {
                'chdir': str(mysite_path),
                'app': 'myapp.app_hello_world',
            },
        },
    )
    config_toml_path.write_text(config_toml_contents)
    app = get_app(approot_path)
    html, status, headers, err = run_app(app)
    assert err == []
    assert re.search(r'Hello, World', html)
    assert status == '200 OK'
    assert len(headers) == 2


def test_main_normal_dir_chdir(tmp_path_to_scaffold, passenger_block):
    site_dir_path = tmp_path_to_scaffold
    mysite_path = site_dir_path / 'mysite'
    mysite_path.mkdir()
    mypackage_path = mysite_path / 'mypackage'
    mypackage_path.mkdir()
    init_path = mypackage_path / '__init__.py'
    init_path.touch()
    wsgi_py_path = mypackage_path / 'wsgi.py'
    wsgi_py_path.write_text(
        textwrap.dedent("""
            from wsgi_shim.hello_world import app_hello_world
            application = app_hello_world
        """),
    )
    approot_path = site_dir_path / 'www-approot'
    config_toml_path = approot_path / 'config.toml'
    config_toml_pre_contents = textwrap.dedent(f"""\
        [passenger]
        {{passenger_block}}
        [wsgi]
        chdir = "{mysite_path}"
        app = "mypackage.wsgi.application"
        """)
    config_toml_contents = config_toml_pre_contents.format(passenger_block=passenger_block)
    config_toml_path.write_text(config_toml_contents)

    app = get_app(approot_path)
    html, status, headers, err = run_app(app)
    assert err == []
    assert re.search(r'Hello, World', html)
    assert status == '200 OK'
    assert len(headers) == 2


def test_main_normal_chdir_default(tmp_path, passenger_block):
    config_file = tmp_path / 'config.toml'
    config_file.write_text(f"""
    [passenger]
    {passenger_block}
    [wsgi]
    app = "wsgi_shim.hello_world.app_hello_world"
    """)
    app = get_app(tmp_path)
    html, status, headers, err = run_app(app)
    assert err == []
    assert re.search(r'Hello, World', html)
    assert status == '200 OK'
    assert len(headers) == 2


def test_cli_parse_help(capsys):
    with pytest.raises(SystemExit) as exc_info:
        cli_parse(['wsgi-shim', '--help'])
    assert exc_info.value.code == 0
    captured = capsys.readouterr()
    assert 'usage:' in captured.out


def test_cli_parse_bad_argv_count(capsys):
    with pytest.raises(SystemExit) as exc_info:
        cli_parse(['wsgi-shim'])
    assert exc_info.value.code == 2
    captured = capsys.readouterr()
    assert 'usage:' in captured.err


def test_cli_parse_bad_command(capsys):
    with pytest.raises(SystemExit) as exc_info:
        cli_parse(['wsgi-shim', 'bad_command', 'bad_option'])
    assert exc_info.value.code == 2
    captured = capsys.readouterr()
    assert 'usage:' in captured.err


def test_cli_parse_good_command(capsys):
    args = cli_parse(['wsgi-shim', 'install', '/dir/path'])
    assert args.command == 'install'
    assert args.approot_path == Path('/dir/path/www-approot')


def test_cli_backend_install(tmp_path_world_readable, capsys):
    site_dir_path = tmp_path_world_readable
    cli_backend(
        [
            'wsgi-shim',
            'install',
            '--verbose',
            str(site_dir_path),
        ],
    )
    captured = capsys.readouterr()
    assert 'Success' in captured.out
    docroot_path = site_dir_path / 'www-docroot'
    assert docroot_path.is_dir()
    assert (docroot_path.stat().st_mode & 0o007) == 0o005
    approot_path = site_dir_path / 'www-approot'
    assert approot_path.is_dir()
    assert (approot_path.stat().st_mode & 0o007) == 0o005
    restart_dir_path = approot_path / 'tmp'
    assert restart_dir_path.is_dir()
    assert (restart_dir_path.stat().st_mode & 0o007) == 0o005
    restart_txt_path = restart_dir_path / 'restart.txt'
    assert restart_txt_path.is_file()
    assert (restart_txt_path.stat().st_mode & 0o007) == 0o000
    maint_txt_path = restart_dir_path / 'maint.txt'
    assert maint_txt_path.is_file()
    assert (maint_txt_path.stat().st_mode & 0o007) == 0o000


def test_cli_backend_check(tmp_path_world_readable, capsys):
    cli_backend(
        [
            'wsgi-shim',
            'install',
            '--verbose',
            str(tmp_path_world_readable),
        ],
    )
    captured = capsys.readouterr()
    assert 'Success' in captured.out
    cli_backend(
        [
            'wsgi-shim',
            'check',
            '--verbose',
            str(tmp_path_world_readable),
        ],
    )
    captured = capsys.readouterr()
    assert 'Check passed' in captured.out


def test_install_passenger_wsgi_py_normal_maint(tmp_path_world_readable):
    site_dir_path = tmp_path_world_readable
    cli_backend(
        [
            'wsgi-shim',
            'install',
            str(site_dir_path),
        ],
    )
    app_root_path = site_dir_path / 'www-approot'
    html, status, headers, err = run_passenger_wsgi_py(app_root_path)
    assert err == []
    assert re.search(r'put it into maintenance mode', html)
    assert status == '503 Service Unavailable'
    assert len(headers) == 3


def test_install_passenger_wsgi_py_normal_default(tmp_path_world_readable):
    site_dir_path = tmp_path_world_readable
    cli_backend(
        [
            'wsgi-shim',
            'install',
            str(site_dir_path),
        ],
    )
    app_root_path = site_dir_path / 'www-approot'
    maint_path = app_root_path / 'tmp' / 'maint.txt'
    maint_path.unlink()
    html, status, headers, err = run_passenger_wsgi_py(app_root_path)
    assert err == []
    assert re.search(r'Hello, World', html)
    assert status == '200 OK'
    assert len(headers) == 2



import re

from wsgi_shim.wsgi_shim import cli_backend

from tests.conftest import run_passenger_wsgi_py


def test_install_passenger_wsgi_py_normal_flask(
        tmp_path_world_readable,
        passenger_block,
):
    site_dir_path = tmp_path_world_readable
    cli_backend(
        [
            'wsgi-shim',
            'install',
            str(site_dir_path),
        ],
    )
    log_file_path = site_dir_path / 'logfile'
    passenger_app_root_path = site_dir_path / 'www-approot'
    config_toml_path = passenger_app_root_path / 'config.toml'
    config_toml_path.write_text(f"""
    [passenger]
    {passenger_block}
    [wsgi]
    app = "tests.flask_example.app"
    [environment]
    LOG_FILENAME = "{log_file_path}"
    """)
    restart_dir_path = passenger_app_root_path / 'tmp'
    maint_txt_path = restart_dir_path / 'maint.txt'
    maint_txt_path.unlink()
    html, status, headers, err = run_passenger_wsgi_py(passenger_app_root_path)
    assert err == []
    assert re.search(r'Hello, World', html)
    assert status == '200 OK'
    assert len(headers) == 2
    log = log_file_path.read_text()
    assert "INFO tests.flask_example MainThread : Request: /" in log



from .wsgi_shim import check_path_is_not_world_readable
from .wsgi_shim import check_path_is_world_readable
from .wsgi_shim import Config
from .wsgi_shim import get_app
from .wsgi_shim import is_path_world_readable
from .wsgi_shim import load_config
from .wsgi_shim import WSGIConfigException
